import pytest
from pydantic import BaseModel
from tinylang.tools.tool import Tool, process_tools


def sample_function(x: int, y: int = 0) -> int:
    """Sample function for Tool class"""
    return x + y


class SampleInputModel(BaseModel):
    x: int
    y: int = 0


def test_tool_creation():
    tool = Tool(
        name="sample_tool", description="A sample tool", function=sample_function
    )
    assert tool.name == "sample_tool"
    assert tool.description == "A sample tool"
    assert tool.function == sample_function


def test_tool_to_dict():
    tool = Tool(
        name="sample_tool", description="A sample tool", function=sample_function
    )
    tool_dict = tool.to_dict()
    assert tool_dict["type"] == "function"
    assert tool_dict["function"]["name"] == "sample_tool"
    assert tool_dict["function"]["description"] == "A sample tool"
    assert "parameters" in tool_dict["function"]


def test_tool_with_input_model():
    tool = Tool(
        name="sample_tool",
        description="A sample tool",
        function=sample_function,
        input_model=SampleInputModel,
    )
    tool_dict = tool.to_dict()
    assert tool_dict["function"]["parameters"]["properties"]["x"]["type"] == "integer"
    assert tool_dict["function"]["parameters"]["properties"]["y"]["type"] == "integer"
    assert tool_dict["function"]["parameters"]["required"] == ["x"]


def test_process_tools():
    tools = [
        Tool(name="tool1", description="Tool 1", function=lambda x: x),
        Tool(name="tool2", description="Tool 2", function=lambda x, y: x + y),
    ]
    processed_tools = process_tools(tools)
    assert len(processed_tools) == 2
    assert all(tool["type"] == "function" for tool in processed_tools)
    assert processed_tools[0]["function"]["name"] == "tool1"
    assert processed_tools[1]["function"]["name"] == "tool2"


def test_tool_with_complex_input_model():
    class ComplexInputModel(BaseModel):
        name: str
        age: int
        is_student: bool = False
        grades: list[float] = []

    def complex_function(
        name: str, age: int, is_student: bool = False, grades: list[float] = []
    ) -> str:
        return f"Name: {name}, Age: {age}, Student: {is_student}, Grades: {grades}"

    tool = Tool(
        name="complex_tool",
        description="A complex tool",
        function=complex_function,
        input_model=ComplexInputModel,
    )
    tool_dict = tool.to_dict()
    params = tool_dict["function"]["parameters"]
    assert params["properties"]["name"]["type"] == "string"
    assert params["properties"]["age"]["type"] == "integer"
    assert params["properties"]["is_student"]["type"] == "boolean"
    assert params["properties"]["grades"]["type"] == "array"
    assert params["properties"]["grades"]["items"]["type"] == "number"
    assert set(params["required"]) == {"name", "age"}


if __name__ == "__main__":
    pytest.main()



from typing import List, Dict, Optional


class ChatHistory:
    def __init__(
        self,
        max_history: int,
        system_message: str,
        previous_history: Optional[List[Dict[str, str]]] = None,
    ):
        self.max_history = max_history
        self.system_message = system_message
        self.messages: List[Dict[str, str]] = []
        if previous_history:
            for message in previous_history:
                if message["role"] != "system":
                    self.add_message(message["role"], message["content"])

    def add_message(self, role: str, content: str) -> None:
        self.messages.append({"role": role, "content": content})

        # TODO - change this to user-assistant roles and not the entire list
        if self.max_history > 0:
            # Keep only the last max_history * 2 messages
            self.messages = self.messages[-(self.max_history * 2) :]

    def get_messages(self) -> List[Dict[str, str]]:
        return [{"role": "system", "content": self.system_message}] + self.messages

    def clear(self) -> None:
        self.messages.clear()



from typing import Dict, Iterator, AsyncIterable, List, Optional
from .util import get_api_key
from .base import ChatBase
from ..history import ChatHistory
import google.generativeai as genai


class ChatGemini(ChatBase):
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        system_message: str | None = None,
        chat_history: int = 0,
        previous_history: Optional[List[Dict[str, str]]] = None,
        init_kwargs: Dict = {},
    ) -> None:
        self.model = model
        self.api_key = api_key
        self.system_message = system_message or "You are a helpful assistant."
        self.init_kwargs = init_kwargs
        self.chat_history = ChatHistory(
            chat_history, self.system_message, previous_history
        )

        api_key = get_api_key(api_key, "GOOGLE_API_KEY")
        genai.configure(api_key=api_key)

        self.client = genai.GenerativeModel(model, **init_kwargs)
        self.chat_session = self.client.start_chat(
            history=self._convert_history_to_gemini_format()
        )

    def _convert_history_to_gemini_format(self) -> List[Dict[str, str]]:
        gemini_history = []
        for message in self.chat_history.get_messages():
            role = (
                "model"
                if message["role"] == "assistant" or message["role"] == "system"
                else message["role"]
            )
            gemini_history.append({"role": role, "parts": [message["content"]]})
        return gemini_history

    def _update_chat_history(self, role: str, content: str):
        self.chat_history.add_message(role, content)

    def invoke(self, prompt: str) -> str:
        response = self.chat_session.send_message(prompt)
        self.chat_history.add_message("user", prompt)
        content = response.text or ""
        self._update_chat_history("model", content)
        return content

    async def ainvoke(self, prompt: str) -> str:
        response = await self.chat_session.send_message_async(prompt)
        self.chat_history.add_message("user", prompt)
        content = response.text or ""
        self._update_chat_history("model", content)
        return content

    def stream_invoke(self, prompt: str) -> Iterator[str]:
        response = self.chat_session.send_message(prompt, stream=True)
        self.chat_history.add_message("user", prompt)
        full_response = ""
        for chunk in response:
            content = chunk.text or ""
            full_response += content
            yield content
        self._update_chat_history("model", full_response)

    async def astream_invoke(self, prompt: str) -> AsyncIterable[str]:
        response = await self.chat_session.send_message_async(prompt, stream=True)
        self.chat_history.add_message("user", prompt)
        full_response = ""
        async for chunk in response:
            content = chunk.text or ""
            full_response += content
            yield content
        self._update_chat_history("model", full_response)

    def get_history(self) -> List[Dict[str, str]]:
        return self.chat_history.get_messages()



from typing import List, Dict, Any, Callable, Optional, Type, Union
from pydantic import BaseModel, create_model
from pydantic.json_schema import model_json_schema
import inspect


class Tool:
    def __init__(
        self,
        name: str,
        description: str,
        function: Union[Callable, Callable[..., Any]],
        input_model: Optional[Type[BaseModel]] = None,
    ):
        self.name = name
        self.description = description
        self.function = function
        self.input_model = input_model or self._create_input_model()
        self.is_async = inspect.iscoroutinefunction(function)

    def to_dict(self) -> Dict[str, Any]:
        schema = model_json_schema(self.input_model)
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": schema,
            },
        }

    def _create_input_model(self) -> Type[BaseModel]:
        annotations = self.function.__annotations__
        fields = {
            name: (annotation, ...)  # ... means the field is required
            for name, annotation in annotations.items()
            if name != "return"
        }
        return create_model(f"{self.name}Input", **fields)


def process_tools(tools: List[Tool]) -> List[Dict[str, Any]]:
    return [tool.to_dict() for tool in tools]



from typing import Iterator, Dict, AsyncIterable, List, Any, Optional
from abc import ABC, abstractmethod


class ChatBase(ABC):
    @abstractmethod
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        system_message: str | None = None,
        init_kwargs: Dict = {},
        chat_history: int = 0,
        previous_history: Optional[List[Dict[str, str]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[str] = "auto",
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def invoke(self, user_input: str) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    async def ainvoke(self, user_input: str) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def stream_invoke(self, user_input: str) -> Iterator[Dict[str, Any]]:
        raise NotImplementedError

    @abstractmethod
    async def astream_invoke(self, user_input: str) -> AsyncIterable[Dict[str, Any]]:
        raise NotImplementedError

    @abstractmethod
    def get_history(self) -> List[Dict[str, str]]:
        raise NotImplementedError

    @abstractmethod
    def clear_history(self) -> None:
        raise NotImplementedError



import os


def get_api_key(api_key: str | None, os_variable: str) -> str:
    if not api_key and not os.getenv(os_variable):
        raise ValueError(
            f"{os_variable} is required, either pass it as an argument or set it as an environment variable as {os_variable}=YOUR_API_KEY"
        )

    return api_key or os.getenv(os_variable)  # type: ignore



from setuptools import find_packages, setup  # type: ignore

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

version = "4.3.0"

setup(
    name="tinylang",
    version=version,
    install_requires=requirements,
    packages=find_packages(),
    description="A tiny language interpreter",
    long_description=long_description,
    long_description_content_type="text/markdown",
)



from typing import Dict, List, Any, Optional, Iterator, AsyncIterable, AsyncIterator
import json
from openai import OpenAI, AsyncOpenAI
from openai.types.chat import ChatCompletion
from .base import ChatBase
from .util import get_api_key
from ..history import ChatHistory
from ..tools.tool import Tool, process_tools


class ChatOpenAI(ChatBase):
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        init_kwargs: Dict = {},
        system_message: str | None = None,
        chat_history: int = 0,
        previous_history: Optional[List[Dict[str, str]]] = None,
        tools: Optional[List[Tool]] = None,
        tool_choice: str | Dict = "auto",
    ) -> None:
        api_key = get_api_key(api_key, "OPENAI_API_KEY")
        init_kwargs.update({"api_key": api_key})
        self.client = OpenAI(**init_kwargs)
        self.async_client = AsyncOpenAI(**init_kwargs)
        self.model = model
        self.system_message = system_message or "You are a helpful assistant."
        self.chat_history = ChatHistory(
            chat_history, self.system_message, previous_history
        )
        self.original_tools = tools
        self.processed_tools = process_tools(tools) if tools else None
        self.tool_choice = tool_choice

    def get_tool_function(self, function_name: str):
        if not self.original_tools:
            raise ValueError("No tools have been set for this chat instance")
        for tool in self.original_tools:
            if tool.name == function_name:
                return tool.function, tool.is_async
        raise ValueError(f"Function {function_name} not found in tools")

    def invoke(self, user_input: Optional[str] = None) -> str:
        if user_input:
            self.chat_history.add_message("user", user_input)
        messages = self.chat_history.get_messages()
        response = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
        )
        return self._process_response(response)

    async def ainvoke(self, user_input: Optional[str] = None) -> str:
        if user_input:
            self.chat_history.add_message("user", user_input)
        messages = self.chat_history.get_messages()
        response = await self.async_client.chat.completions.create(
            model=self.model,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
        )
        return await self._aprocess_response(response)

    def stream_invoke(self, user_input: Optional[str] = None) -> Iterator[str]:
        if user_input:
            self.chat_history.add_message("user", user_input)
        messages = self.chat_history.get_messages()
        response = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
            stream=True,
        )
        yield from self._stream_process_response(response)

    async def astream_invoke(
        self, user_input: Optional[str] = None
    ) -> AsyncIterator[str]:
        if user_input:
            self.chat_history.add_message("user", user_input)
        messages = self.chat_history.get_messages()
        response = await self.async_client.chat.completions.create(
            model=self.model,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
            stream=True,
        )
        async for chunk in self._astream_process_response(response):
            yield chunk

    def _process_response(self, response: ChatCompletion) -> str:
        message = response.choices[0].message
        if message.tool_calls:
            self.chat_history.messages.append(message)

            for tool_call in message.tool_calls:
                function_name = tool_call.function.name
                function_args = json.loads(tool_call.function.arguments)

                print(f"Calling function {function_name} with args {function_args}")
                function_to_call, is_async = self.get_tool_function(function_name)
                if is_async:
                    raise ValueError(
                        "Async tools cannot be used with synchronous methods. Use the async methods instead such as .ainvoke()."
                    )
                result = function_to_call(**function_args)

                self.chat_history.messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "name": function_name,
                        "content": str(result),
                    }
                )

            return self.invoke()
        else:
            content = message.content
            self.chat_history.add_message("assistant", content)
            return content

    def _stream_process_response(
        self, response: Iterator[ChatCompletion]
    ) -> Iterator[str]:
        tool_calls = []
        full_content = ""

        for chunk in response:
            if not chunk.choices:
                continue

            delta = chunk.choices[0].delta
            if delta.content:
                full_content += delta.content
                yield delta.content

            if delta.tool_calls:
                for tc_chunk in delta.tool_calls:
                    if len(tool_calls) <= tc_chunk.index:
                        tool_calls.append(
                            {
                                "id": "",
                                "type": "function",
                                "function": {"name": "", "arguments": ""},
                            }
                        )
                    tc = tool_calls[tc_chunk.index]
                    if tc_chunk.id:
                        tc["id"] += tc_chunk.id
                    if tc_chunk.function and tc_chunk.function.name:
                        tc["function"]["name"] += tc_chunk.function.name
                    if tc_chunk.function and tc_chunk.function.arguments:
                        tc["function"]["arguments"] += tc_chunk.function.arguments

        if tool_calls:
            self.chat_history.messages.append(
                {"role": "assistant", "content": None, "tool_calls": tool_calls}
            )

            for tool_call in tool_calls:
                function_name = tool_call["function"]["name"]
                function_args = json.loads(tool_call["function"]["arguments"])
                function_to_call, is_async = self.get_tool_function(function_name)
                if is_async:
                    raise ValueError(
                        "Async tools cannot be used with synchronous methods. Use the async methods instead such as .astream_invoke()."
                    )
                result = function_to_call(**function_args)

                self.chat_history.messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call["id"],
                        "name": function_name,
                        "content": str(result),
                    }
                )

            # Recursively call stream_invoke() without passing any user input
            yield from self.stream_invoke()
        else:
            self.chat_history.add_message("assistant", full_content)

    async def _astream_process_response(
        self, response: AsyncIterator[ChatCompletion]
    ) -> AsyncIterator[str]:
        tool_calls = []
        full_content = ""

        async for chunk in response:
            if not chunk.choices:
                continue

            delta = chunk.choices[0].delta
            if delta.content:
                full_content += delta.content
                yield delta.content

            if delta.tool_calls:
                for tc_chunk in delta.tool_calls:
                    if len(tool_calls) <= tc_chunk.index:
                        tool_calls.append(
                            {
                                "id": "",
                                "type": "function",
                                "function": {"name": "", "arguments": ""},
                            }
                        )
                    tc = tool_calls[tc_chunk.index]
                    if tc_chunk.id:
                        tc["id"] += tc_chunk.id
                    if tc_chunk.function and tc_chunk.function.name:
                        tc["function"]["name"] += tc_chunk.function.name
                    if tc_chunk.function and tc_chunk.function.arguments:
                        tc["function"]["arguments"] += tc_chunk.function.arguments

        if tool_calls:
            self.chat_history.messages.append(
                {"role": "assistant", "content": None, "tool_calls": tool_calls}
            )

            for tool_call in tool_calls:
                function_name = tool_call["function"]["name"]
                function_args = json.loads(tool_call["function"]["arguments"])
                function_to_call, is_async = self.get_tool_function(function_name)
                print(f"Calling function {function_name} with args {function_args}")
                if is_async:
                    result = await function_to_call(**function_args)
                else:
                    result = function_to_call(**function_args)

                self.chat_history.messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call["id"],
                        "name": function_name,
                        "content": str(result),
                    }
                )

            # Recursively call astream_invoke() without passing any user input
            async for chunk in self.astream_invoke():
                yield chunk
        else:
            self.chat_history.add_message("assistant", full_content)

    async def _aprocess_response(self, response: ChatCompletion) -> str:
        message = response.choices[0].message
        if message.tool_calls:
            self.chat_history.messages.append(message)

            for tool_call in message.tool_calls:
                function_name = tool_call.function.name
                function_args = json.loads(tool_call.function.arguments)

                print(f"Calling function {function_name} with args {function_args}")
                try:
                    function_to_call, is_async = self.get_tool_function(function_name)
                    if is_async:
                        result = await function_to_call(**function_args)
                    else:
                        return function_to_call(**function_args)
                except ValueError:
                    result = f"Function '{function_name}' is not implemented."

                self.chat_history.messages.append(
                    {
                        "role": "tool",
                        "tool_call_id": tool_call.id,
                        "name": function_name,
                        "content": str(result),
                    }
                )

            # Recursively call ainvoke() without passing any user input
            return await self.ainvoke()
        else:
            content = message.content
            self.chat_history.add_message("assistant", content)
            return content

    def get_history(self) -> List[Dict[str, str]]:
        return self.chat_history.get_messages()

    def clear_history(self) -> None:
        self.chat_history.clear()



from typing import Dict, Iterator, AsyncIterable, List, Optional, Any, AsyncIterator
import re
from .base import ChatBase
import anthropic
from anthropic import AsyncAnthropic
from anthropic.types import ToolUseBlock
from .util import get_api_key
from ..history import ChatHistory
from ..tools.tool import Tool, process_tools


class ChatClaude(ChatBase):
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        init_kwargs: Dict = {},
        system_message: str | None = None,
        chat_history: int = 0,
        previous_history: Optional[List[Dict[str, str]]] = None,
        tools: Optional[List[Tool]] = None,
        tool_choice: str | Dict = "auto",
    ) -> None:
        api_key = get_api_key(api_key, "ANTHROPIC_API_KEY")
        init_kwargs.update({"api_key": api_key})
        self.client = anthropic.Anthropic(**init_kwargs)
        self.async_client = AsyncAnthropic(**init_kwargs)
        self.model = model
        self.system_message = system_message or "You are a helpful assistant."
        self.chat_history = ChatHistory(
            chat_history, self.system_message, previous_history
        )
        self.original_tools = tools
        self.processed_tools = self._process_tools(tools)
        self.tool_choice = tool_choice

    def _process_tools(
        self, tools: Optional[List[Tool]]
    ) -> Optional[List[Dict[str, Any]]]:
        """
        Process the provided tools and format them according to Claude's specifications.

        Args:
            tools (Optional[List[Tool]]): A list of Tool objects to be processed, or None.

        Returns:
            Optional[List[Dict[str, Any]]]: A list of dictionaries representing the processed tools, or None if no tools were provided.
        """
        if not tools:
            return None

        processed_tools = []
        for tool in tools:
            processed_tool = {
                "name": tool.name,
                "description": tool.description,
                "input_schema": {"type": "object", "properties": {}, "required": []},
            }

            # Ensure the tool name matches Claude's regex requirement
            if not re.match(r"^[a-zA-Z0-9_-]{1,64}$", tool.name):
                raise ValueError(
                    f"Tool name '{tool.name}' does not match the required format."
                )

            # Process input schema
            tool_dict = tool.to_dict()
            for param_name, param_info in tool_dict["function"]["parameters"][
                "properties"
            ].items():
                processed_tool["input_schema"]["properties"][param_name] = {
                    "type": param_info["type"],
                    "description": param_info.get("description", ""),
                }

                # Handle enums
                if "enum" in param_info:
                    processed_tool["input_schema"]["properties"][param_name]["enum"] = (
                        param_info["enum"]
                    )

                # Add to required list if the field is required
                if param_name in tool_dict["function"]["parameters"].get(
                    "required", []
                ):
                    processed_tool["input_schema"]["required"].append(param_name)

            processed_tools.append(processed_tool)

        return processed_tools

    def get_tool_function(self, function_name: str):
        if not self.original_tools:
            raise ValueError("No tools provided.")

        for tool in self.original_tools:
            if tool.name == function_name:
                return tool.function, tool.is_async

        raise ValueError(f"Tool {function_name} not found.")

    def invoke(self, prompt: Optional[str] = None, max_tokens: int = 2048) -> str:
        if prompt is not None:
            self.chat_history.add_message("user", prompt)
        messages = self.chat_history.get_messages()[1:]  # Exclude system message
        response = self.client.messages.create(
            model=self.model,
            system=self.system_message,
            max_tokens=max_tokens,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
        )
        return self._process_response(response)

    async def ainvoke(self, prompt: Optional[str] = None) -> str:
        if prompt is not None:
            self.chat_history.add_message("user", prompt)
        messages = self.chat_history.get_messages()[1:]  # Exclude system message
        response = await self.async_client.messages.create(
            model=self.model,
            system=self.system_message,
            max_tokens=2048,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
        )
        return await self._process_async_response(response)

    def stream_invoke(self, prompt: Optional[str] = None) -> Iterator[str]:
        if prompt is not None:
            self.chat_history.add_message("user", prompt)
        messages = self.chat_history.get_messages()[1:]  # Exclude system message
        with self.client.messages.stream(
            model=self.model,
            system=self.system_message,
            max_tokens=2048,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
        ) as stream:
            yield from self._stream_process_response(stream)

    async def astream_invoke(self, prompt: Optional[str] = None) -> AsyncIterable[str]:
        if prompt is not None:
            self.chat_history.add_message("user", prompt)
        messages = self.chat_history.get_messages()[1:]  # Exclude system message
        async with self.async_client.messages.stream(
            model=self.model,
            system=self.system_message,
            max_tokens=2048,
            messages=messages,
            tools=self.processed_tools,
            tool_choice=self.tool_choice,
        ) as stream:
            async for chunk in self._astream_process_response(stream):
                yield chunk

    def _process_response(self, response) -> str:
        if response.stop_reason == "tool_use":
            self.chat_history.add_message("assistant", str(response.content))

            tool_response_content = []

            for res in response.content:
                if res.type != "tool_use":
                    continue

                tool_id = res.id
                tool_name = res.name
                tool_input = res.input

                function, is_async = self.get_tool_function(tool_name)
                if is_async:
                    raise ValueError(
                        "Async tools cannot be used with synchronous methods, use ainvoke() instead."
                    )
                print(f"Calling function {tool_name} with args {tool_input}")
                result = function(**tool_input)

                tool_response = {
                    "type": "tool_result",
                    "tool_use_id": tool_id,
                    "content": str(result),
                }

                tool_response_content.append(tool_response)

            self.chat_history.add_message("user", str(tool_response_content))

            # Recursively call invoke() without passing any user input
            return self.invoke()
        else:
            content = response.content[0].text or ""
            self.chat_history.add_message("assistant", content)
            return content

    async def _process_async_response(self, response) -> str:
        if response.stop_reason == "tool_use":
            self.chat_history.add_message("assistant", str(response.content))

            tool_response_content = []

            for res in response.content:
                if res.type != "tool_use":
                    continue

                tool_id = res.id
                tool_name = res.name
                tool_input = res.input

                function, is_async = self.get_tool_function(tool_name)
                print(f"Calling function {tool_name} with args {tool_input}")
                if is_async:
                    result = await function(**tool_input)
                else:
                    result = function(**tool_input)

                tool_response = {
                    "type": "tool_result",
                    "tool_use_id": tool_id,
                    "content": str(result),
                }

                tool_response_content.append(tool_response)

            self.chat_history.add_message("user", str(tool_response_content))

            # Recursively call ainvoke() without passing any user input
            return await self.ainvoke()
        else:
            content = response.content[0].text or ""
            self.chat_history.add_message("assistant", content)
            return content

    def _stream_process_response(self, stream) -> Iterator[str]:
        full_content = ""
        for event in stream:
            if event.type == "text":
                full_content += event.text
                yield event.text

        final_message = stream.get_final_message()
        self.chat_history.add_message("assistant", full_content)

        tool_calls = [
            block for block in final_message.content if isinstance(block, ToolUseBlock)
        ]

        if tool_calls:
            tool_response_content = []

            for tool_call in tool_calls:
                function_name = tool_call.name
                function_args = tool_call.input

                function_to_call, is_async = self.get_tool_function(function_name)
                if is_async:
                    raise ValueError(
                        "Async tools cannot be used with synchronous methods, use astream_invoke() instead."
                    )
                print(f"Calling function {function_name} with args {function_args}")
                function_response = function_to_call(**function_args)

                tool_result = {
                    "type": "tool_result",
                    "tool_use_id": tool_call.id,
                    "content": str(function_response),
                }

                tool_response_content.append(tool_result)

            self.chat_history.add_message("user", str(tool_response_content))

            # Recursively call stream_invoke() without passing any user input
            yield from self.stream_invoke()

    async def _astream_process_response(self, stream) -> AsyncIterator[str]:
        full_content = ""
        async for event in stream:
            if event.type == "text":
                full_content += event.text
                yield event.text

        final_message = await stream.get_final_message()
        self.chat_history.add_message("assistant", full_content)

        tool_calls = [
            block for block in final_message.content if isinstance(block, ToolUseBlock)
        ]

        if tool_calls:
            tool_response_content = []

            for tool_call in tool_calls:
                function_name = tool_call.name
                function_args = tool_call.input
                function_to_call, is_async = self.get_tool_function(function_name)

                print(f"Calling function {function_name} with args {function_args}")
                if is_async:
                    function_response = await function_to_call(**function_args)
                else:
                    function_response = function_to_call(**function_args)

                tool_result = {
                    "type": "tool_result",
                    "tool_use_id": tool_call.id,
                    "content": str(function_response),
                }

                tool_response_content.append(tool_result)

            self.chat_history.add_message("user", str(tool_response_content))

            # Recursively call astream_invoke() without passing any user input
            async for chunk in self.astream_invoke():
                yield chunk

    def get_history(self) -> List[Dict[str, str]]:
        return self.chat_history.get_messages()

    def clear_history(self) -> None:
        self.chat_history.clear()
